'''
Python module handling simulation logic in Infection Explorer
'''
import random
from infection import Person
from visualizer import draw_person, canvas, draw_buildings, graph_sick_fraction
from browser import document, window
from browser.timer import request_animation_frame
from browser.local_storage import storage
import json
# import time

canvas = document['canvas']
ctx = canvas.getContext('2d')


SIM_WINDOW = ((0, canvas.width), (0, canvas.height))  # (x-x), (y-y)
HOUSEHOLDS = [1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 5, 6]
TICKS_IN_DAY = 99


class Simulation():
    def __init__(self, population_count: int=100):
        self.running = False
        self.population_count = population_count
        self.population = [Person(id+1, age=random.randint(1, 100)) for id in range(population_count)]
        self.households = []
        self.existing_social_distancing = 3
        self.day = 1
        self.tick = 0
        self.generate_households()

        self.contagious_persons = []
        self.graph_points = []

        self.healthy_count = lambda: self._status_count('healthy')
        self.asymptomatic_count = lambda: self._status_count('asymptomatic')
        self.sick_count = lambda: self._status_count('sick')
        self.critical_count = lambda: self._status_count('critical')
        self.recovered_count = lambda: self._status_count('recovered')
        self.dead_count = lambda: self._status_count('dead')
        self.alive_count = lambda: len(self.population) - self.dead_count()
        self.sick_fraction = lambda: (self.sick_count() + self.critical_count()) / self.alive_count()

        # self.t1 = 0.0
        # self.t2 = 0.0
        # self.t3 = 0.0


    def _status_count(self, status: str) -> int:
        return sum(1 for p in self.population if p.condition == status)

    def generate_households(self):
        counts_per_house = []
        while sum(counts_per_house) < self.population_count:
            counts_per_house.append(random.choice(HOUSEHOLDS))
        while sum(counts_per_house) > self.population_count:
            counts_per_house[-1] -= 1
        house_width = (SIM_WINDOW[0][1] - SIM_WINDOW[0][0] - 100) / len(counts_per_house)
        current_person_index = 0
        for i, count in enumerate(counts_per_house):
            house_x, house_y = 10 + i * house_width + house_width / 2, 8
            self.households.append({'x': house_x, 'y': house_y, 'people': self.population[current_person_index:current_person_index+count]})
            current_person_index += count
        for d in self.households:
            x, y = d['x'], d['y']
            for person in d['people']:
                person.home_coords = (x, y)
                person.x = x
                person.y = (y + 2) * 2.5 + 20 * (random.random() - 0.5)

    def draw_all_people(self):
        ctx.clearRect(0, 0, 800, 500)
        draw_buildings()
        for person in self.population:
            draw_person(person.x, person.y, person.condition)

    def write_stats(self):
        ctx.fillStyle = '#000'
        ctx.save()
        ctx.translate(0, 500)
        ctx.scale(1, -1)
        ctx.beginPath()
        ctx.fillText('Alive: ' + str(self.alive_count()), 8, 30)
        ctx.fillText('Dead: ' + str(self.dead_count()), 8, 60)
        ctx.fillText('Day: ' + str(self.day) + '.' + str(self.tick), 320, 30)
        ctx.restore()

    def check_target_locations(self):
        social_distancing = json.loads(storage['social_distancing'])
        if social_distancing != self.existing_social_distancing:
            self.existing_social_distancing = social_distancing
            Person.social_distancing = social_distancing
            for person in self.population:
                person.reset_target_location(home=False)
            print('social_distancing updated to ' + str(social_distancing))

        if self.tick == round(TICKS_IN_DAY * 0.65, 0):
            for person in self.population:
                person.reset_target_location(home=True)


    def update(self, *args):
        self.tick += 1
        self.check_target_locations()

        for person in self.population:
            if random.random() < 0.3:
                person.move(self.tick, self.day)

        if self.tick % 2 == 0:
            self.contagious_persons = [p for p in self.population if p.condition in ['asymptomatic', 'sick', 'critical']]

        for i, person in enumerate(self.population):
            if i % 15 == self.tick % 15:
                person.update_condition(self.population, self.day, self.contagious_persons)

        self.draw_all_people()
        self.write_stats()

        if self.tick % 5 == 0:
            Person.sick_fraction = self.sick_fraction()
            self.graph_points.append((self.day + (self.tick / 100), Person.sick_fraction))
        graph_sick_fraction(self.graph_points)

        if self.tick >= TICKS_IN_DAY:
            self.tick = 1
            self.day += 1
            for person in self.population:
                person.reset_target_location(home=False)

    def run_simulation(self, *args):
        if json.loads(storage['restart_sim']):
            storage['restart_sim'] = json.dumps(False)
            self.day = 1
            self.tick = 0
            print('Simulation restarted')
            for person in self.population:
                person.reset()
                person.reset_target_location(home=False)
            self.contagious_persons = []
            self.graph_points = []
            self.run_simulation()
        else:
            if self.day + self.tick / 100 <= 20.01: # 50
                request_animation_frame(self.run_simulation)
                self.update()
            else:
                pass
                # print(f't1: {self.t1}')
                # print(f't2: {self.t2}')
                # print(f't3: {self.t3}')


sim = Simulation(population_count=100)
sim.draw_all_people()
storage['restart_sim'] = json.dumps(True)

def restart_sim(*args):
    storage['restart_sim'] = json.dumps(True)
    if not sim.running:
        sim.run_simulation()
start_btn = document['start']
start_btn.bind('click', restart_sim)
